
/******************************************************************************
*	Graphic Loader Class Source
******************************************************************************/

#include "stdafx.h"
#include "GraphicLoader.h"
#ifdef SLIB_USE_GRAPHICLOADER
using namespace selna;


/*----------------------------------------------------------------------------*
|	RXgN^
*----------------------------------------------------------------------------*/
CGraphicLoader::CGraphicLoader()
{
	m_hBmpInfo = NULL;
	m_hBmpBits = NULL;
}


/*----------------------------------------------------------------------------*
|	fXgN^
*----------------------------------------------------------------------------*/
CGraphicLoader::~CGraphicLoader()
{
	Release();
}


/*----------------------------------------------------------------------------*
|	
*----------------------------------------------------------------------------*/
void CGraphicLoader::Release()
{
	if( m_hBmpInfo )
	{
		::LocalFree( m_hBmpInfo );
		m_hBmpInfo = NULL;
	}

	if( m_hBmpBits )
	{
		::LocalFree( m_hBmpBits );
		m_hBmpBits = NULL;
	}
}


/*----------------------------------------------------------------------------*
|	TCY擾
*----------------------------------------------------------------------------*/
BOOL CGraphicLoader::GetSize( SIZE* pSize )
{
	if( pSize == NULL ) return FALSE;
	if( m_hBmpInfo == NULL ) return FALSE;

	BITMAPINFO* pBmpInfo = (BITMAPINFO*)::LocalLock( m_hBmpInfo );
	{
		if( !pBmpInfo ) return FALSE;

		pSize->cx = pBmpInfo->bmiHeader.biWidth;
		pSize->cy = pBmpInfo->bmiHeader.biHeight;

	}
	::LocalUnlock( m_hBmpInfo );

	return TRUE;
}


/*----------------------------------------------------------------------------*
|	pbg擾
*----------------------------------------------------------------------------*/
BOOL CGraphicLoader::GetPalette( LPPALETTEENTRY pPal )
{

	if( pPal == NULL ) return FALSE;
	if( m_hBmpInfo == NULL ) return FALSE;

	BITMAPINFO* pBmpInfo = (BITMAPINFO*)::LocalLock( m_hBmpInfo );
	{
		if( !pBmpInfo ) return FALSE;

		if( pBmpInfo->bmiHeader.biBitCount != 8 ) 
		{
			return FALSE;
			::LocalUnlock( m_hBmpInfo );
		}

		for( int i = 0; i < 256; i++ )
		{
			pPal[i].peRed   = pBmpInfo->bmiColors[i].rgbRed;
			pPal[i].peGreen = pBmpInfo->bmiColors[i].rgbGreen;
			pPal[i].peBlue  = pBmpInfo->bmiColors[i].rgbBlue;
			pPal[i].peFlags = pBmpInfo->bmiColors[i].rgbReserved;
		}
	}
	::LocalUnlock( m_hBmpInfo );

	return TRUE;
}


/*----------------------------------------------------------------------------*
|	pbg擾
*----------------------------------------------------------------------------*/
BOOL CGraphicLoader::GetPalette( LPRGBQUAD pPal )
{

	if( pPal == NULL ) return FALSE;
	if( m_hBmpInfo == NULL ) return FALSE;

	BITMAPINFO* pBmpInfo = (BITMAPINFO*)::LocalLock( m_hBmpInfo );
	{
		if( !pBmpInfo ) return FALSE;

		if( pBmpInfo->bmiHeader.biBitCount != 8 ) 
		{
			return FALSE;
			::LocalUnlock( m_hBmpInfo );
		}

		for( int i = 0; i < 256; i++ )
		{
			pPal[i].rgbRed      = pBmpInfo->bmiColors[i].rgbRed;
			pPal[i].rgbGreen    = pBmpInfo->bmiColors[i].rgbGreen;
			pPal[i].rgbBlue     = pBmpInfo->bmiColors[i].rgbBlue;
			pPal[i].rgbReserved = pBmpInfo->bmiColors[i].rgbReserved;
		}
	}
	::LocalUnlock( m_hBmpInfo );

	return TRUE;
}


/*----------------------------------------------------------------------------*
|	sN`[ǂݍ
*----------------------------------------------------------------------------*/
void CGraphicLoader::LoadPicture( const char* lpszFileName, DWORD dwOffset, const char* lpszSpiDir )
{
	try
	{
		CFile fin( lpszFileName, "rb" );

		if( !fin ) throw ExceptionEx( "File Not Found" );

		if( lpszSpiDir == NULL )
		{
			// BITMAP
			fin.Seek( dwOffset, SEEK_SET );
			if( LoadBitmap( fin ) )	return;

			// JPEG
			fin.Seek( dwOffset, SEEK_SET );
			if( LoadJpeg( fin ) )	return;

			// PNG
			fin.Seek( dwOffset, SEEK_SET );
			if( LoadPng( fin ) )	return;

			// TARGA
			fin.Seek( dwOffset, SEEK_SET );
			if( LoadTarga( fin ) )	return;

		}else
		{
			// SPI Plugin
			fin.Seek( dwOffset, SEEK_SET );
			if( LoadBySPI( fin, lpszSpiDir ) ) return;

		}

		throw ExceptionEx( "Picture Format Error" );
	}

	catch( exception& e )
	{
		throw ExceptionEx( "ErrorMsg=%s\nFile=%s\nOffset=%d", e.what(), lpszFileName, dwOffset );
	}

}


/*----------------------------------------------------------------------------*
|	rbg}bvǂݍ
*----------------------------------------------------------------------------*/
bool CGraphicLoader::LoadBitmap( CFile& fin )
{

	BITMAPFILEHEADER bf;
	fin.Read( &bf, sizeof(BITMAPFILEHEADER), 1 );

	if( bf.bfType != 0x4D42 ) return false;

	BITMAPINFOHEADER bi;
	fin.Read( &bi, sizeof(BITMAPINFOHEADER), 1 );

	if( bi.biCompression != BI_RGB ) throw ExceptionEx( "BMP No RGB" );
		
	if( bi.biHeight < 0 ) throw ExceptionEx( "BMP No BottomUp" );

	UINT uInfoSize = bf.bfOffBits-sizeof(BITMAPFILEHEADER);
	UINT uBitsSize = (bf.bfSize-bf.bfOffBits);

	m_hBmpInfo = ::LocalAlloc( LMEM_MOVEABLE, uInfoSize );
	m_hBmpBits = ::LocalAlloc( LMEM_MOVEABLE, uBitsSize );

	// mۃG[
	if( !m_hBmpInfo || !m_hBmpBits ) throw bad_alloc();

	BITMAPINFO* pBmpInfo = (BITMAPINFO*)::LocalLock( m_hBmpInfo );
	BYTE* pBmpBits = (BYTE*)::LocalLock( m_hBmpBits );
	{
		// G[
		if( !pBmpInfo || !pBmpBits ) throw bad_alloc();

		// rbg}bvCtHRs[
		pBmpInfo->bmiHeader = bi;
		
		UINT uTblSize = uInfoSize-sizeof(BITMAPINFOHEADER);

		if( uTblSize != 0 )
		{
			// J[e[uǂݍ
			fin.Read(  &pBmpInfo->bmiColors, sizeof(BYTE), uTblSize );
		}
	
		// sNZf[^ǂݍ
		fin.Read( pBmpBits, sizeof(BYTE), uBitsSize );

	}
	::LocalUnlock( m_hBmpInfo );
	::LocalUnlock( m_hBmpBits );

	return true;
}


/*----------------------------------------------------------------------------*
|	JPEGǂݍ
*----------------------------------------------------------------------------*/
bool CGraphicLoader::LoadJpeg( CFile& fin )
{
	jpeg_decompress_struct cinfo;
	gl_jpeg_error_mgr jerr;

	JSAMPROW buffer[1];
	BITMAPINFO* pBmpInfo = NULL;
	BYTE* pBmpBits = NULL;

	BYTE sig[11]={ 0 };
	fin.Read( sig, sizeof(BYTE), 11 );
	fin.Seek( -11L, SEEK_CUR );

	// JPEGt@C`FbN
	if( (sig[0]!=0xFF) || (sig[1]!=0xD8) ) return false; 
	if( (sig[2]!=0xFF) || (sig[3]!=0xE0) ) return false;
	if ( strcmp((char*)&sig[6],"JFIF") != 0 ) return false;

	// G[̃nhO
	cinfo.err = ::jpeg_std_error( &jerr );

	// G[̌Ăяo֐
	jerr.error_exit = JpegErrorExit;

	// G[
	if( ::setjmp( jerr.setjmp_buffer ) )
	{
        // I
		if( pBmpInfo )
			::LocalUnlock( m_hBmpInfo );

		if( pBmpBits )
			::LocalUnlock( m_hBmpBits );

		::jpeg_destroy_decompress( &cinfo );

		throw ExceptionEx("JPEG Abort");
    }

	// \̂̏ݒ
	::jpeg_create_decompress( &cinfo );

	// t@C̓nh̐ݒ
	::jpeg_stdio_src( &cinfo, fin );

	// t@C̏wb_̓Ǎ
	::jpeg_read_header( &cinfo, TRUE );

	// 𓀊Jn
	::jpeg_start_decompress( &cinfo );

	// sobt@̗̈ݒ
	int row_stride = cinfo.output_width * cinfo.output_components;

	// 4oCgEɏC
	DWORD dwRowBytes = GL_GETBMPROWBYTES( cinfo.image_width, 3 );

	UINT uInfoSize = sizeof(BITMAPINFOHEADER);
	UINT uBitsSize = dwRowBytes*cinfo.image_height;

	m_hBmpInfo = ::LocalAlloc( LMEM_MOVEABLE, uInfoSize );
	m_hBmpBits = ::LocalAlloc( LMEM_MOVEABLE, uBitsSize );

	// mۃG[
	if( !m_hBmpInfo || !m_hBmpBits ) ::longjmp( jerr.setjmp_buffer, true );

	pBmpInfo = (BITMAPINFO*)::LocalLock( m_hBmpInfo );
	pBmpBits = (BYTE*)::LocalLock( m_hBmpBits );
	{
		// G[
		if( !pBmpInfo || !pBmpBits ) ::longjmp( jerr.setjmp_buffer, true );

		// rbg}bvCtHݒ
		pBmpInfo->bmiHeader.biSize			= sizeof( BITMAPINFOHEADER );
		pBmpInfo->bmiHeader.biWidth			= cinfo.image_width;
		pBmpInfo->bmiHeader.biHeight		= cinfo.image_height;
		pBmpInfo->bmiHeader.biPlanes		= 1;
		pBmpInfo->bmiHeader.biBitCount		= 24;
		pBmpInfo->bmiHeader.biCompression	= BI_RGB;
		pBmpInfo->bmiHeader.biSizeImage		= uBitsSize;
		pBmpInfo->bmiHeader.biXPelsPerMeter	= 0; 
		pBmpInfo->bmiHeader.biYPelsPerMeter	= 0; 
		pBmpInfo->bmiHeader.biClrUsed		= 0; 
		pBmpInfo->bmiHeader.biClrImportant	= 0;

		// {gAbvɂ킹邽ߍŉCɂ
		pBmpBits += dwRowBytes*(cinfo.image_height-1);

		BYTE temp;

		DWORD dwPixelByte = cinfo.image_width*3;

		// C[Wǂݍ
		while( cinfo.output_scanline < cinfo.output_height )
		{

			// 1sǂݍ
			buffer[0] = pBmpBits;
			::jpeg_read_scanlines( &cinfo, buffer, 1 );

			// bufferBGRȂ̂RGBɕϊ
			for( int i = 0; i < row_stride; i+=3 )
			{
				temp = pBmpBits[i+0];
				pBmpBits[i+0] = pBmpBits[i+2];
				pBmpBits[i+2] = temp;

			}

			// 4oCgE̎coCg0Ŗ߂Ă
			::memset( &pBmpBits[dwPixelByte], 0, dwRowBytes-dwPixelByte );

			// ̃C
			pBmpBits -= dwRowBytes;
		}

	}
	::LocalUnlock( m_hBmpInfo );
	::LocalUnlock( m_hBmpBits );

	// I
	::jpeg_finish_decompress( &cinfo );
	::jpeg_destroy_decompress( &cinfo );

	return true;
}


/*----------------------------------------------------------------------------*
|	PNGǂݍ
*----------------------------------------------------------------------------*/
bool CGraphicLoader::LoadPng( CFile& fin )
{
	png_structp png_ptr = NULL;
    png_infop  info_ptr = NULL;

	BITMAPINFO* pBmpInfo = NULL;
	BYTE* pBmpBits = NULL;
	auto_array<png_bytep> ppScanLine;

	BYTE sig[8];

	fin.Read( sig, sizeof(BYTE), 8 );

	// PNGt@C`FbN
	if( !::png_check_sig( sig, 8 ) ) return false;

	// PNG \̂̎擾
    png_ptr = ::png_create_read_struct( PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    
	if( !png_ptr )  throw ExceptionEx( "PNG Struct Error" );

	// PNG \̂̎擾
	info_ptr = ::png_create_info_struct( png_ptr );

	if( !info_ptr ) throw ExceptionEx( "PNG Info Struct Error" );

	// G[
	if( ::setjmp( png_ptr->jmpbuf ) )
	{
		// I
		::png_destroy_read_struct( &png_ptr, &info_ptr, NULL);

		throw ExceptionEx( "PNG Abort" );
	}

	// t@Cnhݒ
	::png_init_io( png_ptr, fin );
	
	// wb_̓ǂݍ
	::png_set_sig_bytes( png_ptr, 8 );
	::png_read_info( png_ptr, info_ptr );
	
	png_uint_32 nWidth;
	png_uint_32 nHeight;
	int nBitDepth;
	int nColorType;

	// Kvȏ̎擾
	::png_get_IHDR( png_ptr, info_ptr, &nWidth, &nHeight,
		&nBitDepth, &nColorType, NULL, NULL, NULL);

	switch( nColorType )
	{
	case PNG_COLOR_TYPE_PALETTE:	// 8rbgA16rbg

		if( nBitDepth == 16 ) ::png_set_strip_16( png_ptr );

		nBitDepth = 8;
		break;
	case PNG_COLOR_TYPE_GRAY:		// 8rbgpbgɂ

		nBitDepth = 8;
		break;
	case PNG_COLOR_TYPE_RGB:		// 24rbg RGB

		nBitDepth = 24;
		break;
	case PNG_COLOR_TYPE_RGB_ALPHA:	// 32rbg ARGB

		nBitDepth = 32;
		break;
	default:
		throw ExceptionEx( "PNG Bit Error");
	}

	double  gamma;

	// K}
	if ( ::png_get_gAMA( png_ptr, info_ptr, &gamma) )
	{
		::png_set_gamma( png_ptr, 2.2, gamma );
	}

	// ݒo^
	::png_read_update_info( png_ptr, info_ptr );

	int nPaletteCnt = 0;
	png_color *pPalette;

	if( nColorType == PNG_COLOR_TYPE_PALETTE )
	{
		// pbgǂݍ
		::png_get_PLTE( png_ptr, info_ptr, &pPalette, &nPaletteCnt );

	}else if( nColorType == PNG_COLOR_TYPE_GRAY )
	{
		// pbgݒ
		nPaletteCnt = 256;

	}else
	{
		// RGBBGRɕϊ
		::png_set_bgr( png_ptr );
	}

	// 4oCgEɏC
	DWORD dwRowBytes = GL_GETBMPROWBYTES( nWidth, nBitDepth>>3);
	//(png_get_rowbytes( png_ptr, info_ptr )+3)&~3;

	UINT uInfoSize = sizeof(BITMAPINFOHEADER)+sizeof(RGBQUAD)*nPaletteCnt;
	UINT uBitsSize = dwRowBytes*nHeight;

	m_hBmpInfo = ::LocalAlloc( LMEM_MOVEABLE, uInfoSize );
	m_hBmpBits = ::LocalAlloc( LMEM_MOVEABLE, uBitsSize );

	// mۃG[
	if( !m_hBmpInfo || !m_hBmpBits ) ::longjmp( png_ptr->jmpbuf, true );

	pBmpInfo = (BITMAPINFO*)::LocalLock( m_hBmpInfo );
	pBmpBits = (BYTE*)::LocalLock( m_hBmpBits );
	{
		// G[
		if( !pBmpInfo || !pBmpBits ) ::longjmp( png_ptr->jmpbuf, true );

		// rbg}bvCtHݒ
		pBmpInfo->bmiHeader.biSize			= sizeof( BITMAPINFOHEADER );
		pBmpInfo->bmiHeader.biWidth			= nWidth;
		pBmpInfo->bmiHeader.biHeight		= nHeight;
		pBmpInfo->bmiHeader.biPlanes		= 1;
		pBmpInfo->bmiHeader.biBitCount		= nBitDepth;
		pBmpInfo->bmiHeader.biCompression	= BI_RGB;
		pBmpInfo->bmiHeader.biSizeImage		= uBitsSize;
		pBmpInfo->bmiHeader.biXPelsPerMeter	= 0; 
		pBmpInfo->bmiHeader.biYPelsPerMeter	= 0; 
		pBmpInfo->bmiHeader.biClrUsed		= 0; 
		pBmpInfo->bmiHeader.biClrImportant	= 0;

		if( nPaletteCnt )
		{
			RGBQUAD* pBmiColors = pBmpInfo->bmiColors;

			if( nColorType == PNG_COLOR_TYPE_GRAY )
			{
				// OCXP[pbg쐬
				for( int n = 0; n < nPaletteCnt; n++ )
				{
					pBmiColors[n].rgbRed      = n;
					pBmiColors[n].rgbGreen    = n;
					pBmiColors[n].rgbBlue     = n;
					pBmiColors[n].rgbReserved = 0;
				}

			}else
			{
				// pbgRs[
				for( int n = 0; n < nPaletteCnt; n++ )
				{
					pBmiColors[n].rgbRed      = pPalette[n].red;
					pBmiColors[n].rgbGreen    = pPalette[n].green;
					pBmiColors[n].rgbBlue     = pPalette[n].blue;
					pBmiColors[n].rgbReserved = 0;
				}
			}
		}

		try
		{
			ppScanLine = auto_array<png_bytep>( new png_bytep[ nHeight ] );
		}
		catch( ... )
		{
			::longjmp( png_ptr->jmpbuf, true );
		}

		DWORD dwPixelByte = nWidth*(nBitDepth>>3);

		for( int i = nHeight-1; i >= 0 ; i-- )
		{
			ppScanLine[i] = pBmpBits;

			// 4oCgE̎coCg0Ŗ߂Ă
			::memset( &pBmpBits[dwPixelByte], 0, dwRowBytes-dwPixelByte );

			pBmpBits += dwRowBytes;
		}

		// C[Wǂݍ
		::png_read_image( png_ptr, ppScanLine );
		::png_read_end( png_ptr, NULL );

	}
	::LocalUnlock( m_hBmpInfo );
	::LocalUnlock( m_hBmpBits );

	// I
	::png_destroy_read_struct( &png_ptr, &info_ptr, NULL);
	
	return true;
}


/*----------------------------------------------------------------------------*
|	TARGAǂݍ
*----------------------------------------------------------------------------*/
bool CGraphicLoader::LoadTarga( CFile& fin )
{
	
	TARGAHEADER th;
	fin.Read( &th, sizeof(TARGAHEADER), 1 );
	
	if( th.Image_Type != 2 && th.Image_Type != 1 ) return false;

	if( th.Pixel != 8 && th.Pixel != 24 && th.Pixel != 32 ) return false;
	
	long nBitCount = th.Pixel;

	if( nBitCount == 8 )
	{
		// `ltJ[e[uH
		if( th.CMap_EntrySize == 32 )
		{
			// 32rbgDIBɂ
			nBitCount = 32;
		}
	}

	// No_IDt@C|C^i߂
	fin.Seek( th.No_ID, SEEK_CUR );

	// 4oCgEɏC
	DWORD dwRowBytes = GL_GETBMPROWBYTES( th.Width, nBitCount>>3);

	UINT uInfoSize = sizeof(BITMAPINFOHEADER)+sizeof(RGBQUAD)*((nBitCount==8)?256:0);
	UINT uBitsSize = dwRowBytes*th.Height;

	m_hBmpInfo = ::LocalAlloc( LMEM_MOVEABLE, uInfoSize );
	m_hBmpBits = ::LocalAlloc( LMEM_MOVEABLE, uBitsSize );

	// mۃG[
	if( !m_hBmpInfo || !m_hBmpBits ) throw bad_alloc();
	
	BITMAPINFO* pBmpInfo = (BITMAPINFO*)::LocalLock( m_hBmpInfo );
	BYTE* pBmpBits = (BYTE*)::LocalLock( m_hBmpBits );
	{
		// G[
		if( !pBmpInfo || !pBmpBits ) throw bad_alloc();

		// rbg}bvCtHݒ
		pBmpInfo->bmiHeader.biSize			= sizeof( BITMAPINFOHEADER );
		pBmpInfo->bmiHeader.biWidth			= th.Width;
		pBmpInfo->bmiHeader.biHeight		= th.Height;
		pBmpInfo->bmiHeader.biPlanes		= 1;
		pBmpInfo->bmiHeader.biBitCount		= nBitCount;
		pBmpInfo->bmiHeader.biCompression	= BI_RGB;
		pBmpInfo->bmiHeader.biSizeImage		= uBitsSize;
		pBmpInfo->bmiHeader.biXPelsPerMeter	= 0; 
		pBmpInfo->bmiHeader.biYPelsPerMeter	= 0; 
		pBmpInfo->bmiHeader.biClrUsed		= 0; 
		pBmpInfo->bmiHeader.biClrImportant	= 0;

		switch( th.Pixel )
		{
		case 8:

			if( nBitCount == 32 )
			{
				// `ltJ[e[uǂݍ
				TARGAARGBQUAD APal[256];
				fin.Read( &APal, sizeof(LPTARGAARGBQUAD), 256 );
				
				BYTE Pixel;
				
				int nPixels = th.Width*th.Height;
				LPTARGAARGBQUAD pBits = (LPTARGAARGBQUAD)pBmpBits;

				for( int i = 0; i < nPixels; i++ )
				{
					fin.Read( &Pixel, sizeof(BYTE), 1 );

					pBits[i].alpha = APal[ Pixel ].alpha;
					pBits[i].red   = APal[ Pixel ].red;
					pBits[i].green = APal[ Pixel ].green;
					pBits[i].blue  = APal[ Pixel ].blue;
				}

			}else
			{
				RGBQUAD* pBmiColors = pBmpInfo->bmiColors;

				// J[e[uǂݍ
				for( int n = 0; n < 256; n++ )
				{
					fin.Read( &pBmiColors[n], sizeof(BYTE), 3 );
					pBmiColors[n].rgbReserved = 0;
				}

				for( int h = 0; h < pBmpInfo->bmiHeader.biHeight; h++ )
				{
					fin.Read( pBmpBits, 1, pBmpInfo->bmiHeader.biWidth );

					// 4oCgE̎coCg0Ŗ߂Ă
					::memset( &pBmpBits[pBmpInfo->bmiHeader.biWidth], 0, dwRowBytes-pBmpInfo->bmiHeader.biWidth );

					pBmpBits += dwRowBytes;
				}
			}

			break;
		case 24:
			{
				DWORD dwPixelByte = pBmpInfo->bmiHeader.biWidth*3;

				for( int h = 0; h < pBmpInfo->bmiHeader.biHeight; h++ )
				{
					fin.Read( pBmpBits, 3, pBmpInfo->bmiHeader.biWidth );
					
					// 4oCgE̎coCg0Ŗ߂Ă
					::memset( &pBmpBits[dwPixelByte], 0, dwRowBytes-dwPixelByte );

					pBmpBits += dwRowBytes;
				}
			}
			break;
		case 32:

			fin.Read( pBmpBits, sizeof(BYTE), uBitsSize );
			break;
		}

	}
	::LocalUnlock( m_hBmpInfo );
	::LocalUnlock( m_hBmpBits );

	return true;
}


/*----------------------------------------------------------------------------*
|	SPI Pluginœǂݍ
*----------------------------------------------------------------------------*/
bool CGraphicLoader::LoadBySPI( CFile& fin, const char* lpszSpiDir )
{

	// I[vt@Cݒ
	long nOffset = fin.GetPosition();
	char szImgPath[MAX_PATH];
	strcpy( szImgPath, fin.GetFileName().c_str() );

	// L[hŃt@CI[vȂSPIvOC΍(
	fin.Close();

	// SPIvOCt@CΏۂ̃pX쐬
	char szSearchSpi[ MAX_PATH ];
	::sprintf( szSearchSpi, "%s\\*.spi", lpszSpiDir );

	// t@CT
	WIN32_FIND_DATA fd;
	HANDLE hFind = ::FindFirstFile( szSearchSpi, &fd );

	if( hFind == INVALID_HANDLE_VALUE )
	{
		throw ExceptionEx( "SPI No DLL" );
	}

	HMODULE hDLL = NULL;

	try
	{
		do
		{
			// SPIpX쐬
			char szSpiPath[ MAX_PATH ];
			sprintf( szSpiPath, "%s\\%s", lpszSpiDir, fd.cFileName );

			// SPIvOCǂݍ
			hDLL = ::LoadLibrary( szSpiPath );

			if( hDLL )
			{
				// SPIpɃt@CJ
				HANDLE hFile = ::CreateFile( fin.GetFileName().c_str(),
						GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, 
						NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

				// t@C|C^ړ
				::SetFilePointer( hFile, nOffset, 0, FILE_BEGIN );

				// IsSupported֐擾
				SpiIsSupported IsSupported = (SpiIsSupported)GetProcAddress( hDLL, "IsSupported" );

				int ret = 0;

				if( IsSupported )
				{
					ret = IsSupported( szImgPath, (DWORD)hFile );
				}

				// L[hŃt@CI[vȂSPIvOC΍
				::CloseHandle( hFile );

				if( ret )
				{
					// GetPicture֐擾
					SpiGetPicture GetPicture = (SpiGetPicture)GetProcAddress( hDLL, "GetPicture" );

					if( !GetPicture ) throw ExceptionEx( "SPI DLL Error" );

					// DIBC[W擾
					if( GetPicture( szImgPath, nOffset, 0, &m_hBmpInfo, &m_hBmpBits, NULL, 0 ) )
					{
						throw ExceptionEx( "SPI GetPicture Error" );

					}
				}

				// CuJ
				::FreeLibrary( hDLL );

				if( m_hBmpInfo ) break;
			}

		}while( ::FindNextFile( hFind, &fd ) );


		// I
		::FindClose( hFind );

		if( m_hBmpInfo ) return true;
	}

	catch( ... )
	{

		// I
		if( hDLL ) ::FreeLibrary( hDLL );

		::FindClose( hFind );
		throw;
	}

	return false;
}


/*----------------------------------------------------------------------------*
|	_O
*----------------------------------------------------------------------------*/
BOOL CGraphicLoader::Render( LPVOID lpBuf, DWORD dwDstRowBytes, GLFORMAT ConvFormat )
{
	if( m_hBmpInfo == NULL ) return FALSE;
	if( lpBuf == NULL ) return FALSE;

	int ret = true;

	BITMAPINFO* pBmpInfo = (BITMAPINFO*)::LocalLock( m_hBmpInfo );
	BYTE* pBmpBits = (BYTE*)::LocalLock( m_hBmpBits );
	{
		if( !pBmpInfo || !pBmpBits ) return FALSE;

		long w = pBmpInfo->bmiHeader.biWidth;
		long h = pBmpInfo->bmiHeader.biHeight;

		long ww;
		long hh;

		// ̃oCg𓾂
		DWORD dwSrcRowBytes = GL_GETBMPROWBYTES( pBmpInfo->bmiHeader.biWidth, 
			pBmpInfo->bmiHeader.biBitCount>>3 );

		BYTE* pDstBits = (BYTE*)lpBuf;

		// ŏICɂ킹
		pBmpBits += dwSrcRowBytes*(pBmpInfo->bmiHeader.biHeight-1);

		switch( ConvFormat )
		{
		case GLFMT_R8G8B8: // 24rbg

			switch( pBmpInfo->bmiHeader.biBitCount )
			{
			case 8:
				{
					RGBQUAD* pPal = pBmpInfo->bmiColors;

					for( hh = 0; hh < h; hh++ )
					{
						int d = 0;

						for( ww = 0; ww < w; ww++ )
						{
							pDstBits[d+0] = pPal[ pBmpBits[ww] ].rgbBlue;
							pDstBits[d+1] = pPal[ pBmpBits[ww] ].rgbGreen;
							pDstBits[d+2] = pPal[ pBmpBits[ww] ].rgbRed;
							d+=3;
						}

						pBmpBits -= dwSrcRowBytes;
						pDstBits += dwDstRowBytes;
					}
				}
				break;
			case 24:

				w*=3;

				for( hh = 0; hh < h; hh++ )
				{
					::memmove( pDstBits, pBmpBits, w );

					pBmpBits -= dwSrcRowBytes;
					pDstBits += dwDstRowBytes;
				}

				break;
			case 32:

				for( hh = 0; hh < h; hh++ )
				{
					int s = 0;
					int d = 0;

					for( ww = 0; ww < w; ww++ )
					{
						pDstBits[d+0] = pBmpBits[s+0];
						pDstBits[d+1] = pBmpBits[s+1];
						pDstBits[d+2] = pBmpBits[s+2];
						d+=3;
						s+=4;
					}

					pBmpBits -= dwSrcRowBytes;
					pDstBits += dwDstRowBytes;
				}

				break;
			default:

				ret = false;
			}
			break;
		case GLFMT_X8R8G8B8:	// 32rbg
		case GLFMT_A8R8G8B8:
			{
				BYTE Alpha = ConvFormat==GLFMT_A8R8G8B8?255:0;

				switch( pBmpInfo->bmiHeader.biBitCount )
				{
				case 8:
					{
						RGBQUAD* pPal = pBmpInfo->bmiColors;

						for( hh = 0; hh < h; hh++ )
						{
							int d = 0;

							for( ww = 0; ww < w; ww++ )
							{
								pDstBits[d+0] = pPal[ pBmpBits[ww] ].rgbBlue;
								pDstBits[d+1] = pPal[ pBmpBits[ww] ].rgbGreen;
								pDstBits[d+2] = pPal[ pBmpBits[ww] ].rgbRed;
								pDstBits[d+3] = Alpha;
								d+=4;
							}

							pBmpBits -= dwSrcRowBytes;
							pDstBits += dwDstRowBytes;
						}
					}
					break;
				case 24:

					for( hh = 0; hh < h; hh++ )
					{
						int s = 0;
						int d = 0;

						for( ww = 0; ww < w; ww++ )
						{
							pDstBits[d+0] = pBmpBits[s+0];	// B
							pDstBits[d+1] = pBmpBits[s+1];	// G
							pDstBits[d+2] = pBmpBits[s+2];	// R
							pDstBits[d+3] = Alpha;			// A
							d+=4;
							s+=3;
						}

						pBmpBits -= dwSrcRowBytes;
						pDstBits += dwDstRowBytes;
					}

					break;
				case 32:

					w*=4;

					for( hh = 0; hh < h; hh++ )
					{
						::memmove( pDstBits, pBmpBits, w );

						pBmpBits -= dwSrcRowBytes;
						pDstBits += dwDstRowBytes;
					}

					break;
				default:

					ret = false;
				}
			}
			break;
		case GLFMT_R5G6B5:	// 16rbg 565
			
			switch( pBmpInfo->bmiHeader.biBitCount )
			{
			case 8:
				{
					RGBQUAD* pPal = pBmpInfo->bmiColors;

					for( hh = 0; hh < h; hh++ )
					{
						WORD* pDstBits2 = (WORD*)pDstBits;

						for( ww = 0; ww < w; ww++ )
						{
							pDstBits2[ww] = GL_R8G8B8ToR5G6B5( pPal[pBmpBits[ww]].rgbRed, pPal[pBmpBits[ww]].rgbGreen, pPal[pBmpBits[ww]].rgbBlue );
						}

						pBmpBits -= dwSrcRowBytes;
						pDstBits += dwDstRowBytes;
					}
				}
				break;
			case 24:

				for( hh = 0; hh < h; hh++ )
				{
					int s = 0;
					WORD* pDstBits2 = (WORD*)pDstBits;

					for( ww = 0; ww < w; ww++ )
					{
						pDstBits2[ww] = GL_R8G8B8ToR5G6B5( pBmpBits[s+2], pBmpBits[s+1], pBmpBits[s+0] );
						s+=3;
					}

					pBmpBits -= dwSrcRowBytes;
					pDstBits += dwDstRowBytes;
				}

				break;
			case 32:

				for( hh = 0; hh < h; hh++ )
				{
					int s = 0;
					WORD* pDstBits2 = (WORD*)pDstBits;

					for( ww = 0; ww < w; ww++ )
					{
						pDstBits2[ww] = GL_R8G8B8ToR5G6B5( pBmpBits[s+2], pBmpBits[s+1], pBmpBits[s+0] );
						s+=4;
					}

					pBmpBits -= dwSrcRowBytes;
					pDstBits += dwDstRowBytes;
				}

				break;
			default:

				ret = false;
			}
			break;
		case GLFMT_X1R5G5B5:	// 16rbg 555
		case GLFMT_A1R5G5B5:
			{
				BYTE Alpha = ConvFormat==GLFMT_A1R5G5B5?0xFF:0;

				switch( pBmpInfo->bmiHeader.biBitCount )
				{
				case 8:
					{
						RGBQUAD* pPal = pBmpInfo->bmiColors;

						for( hh = 0; hh < h; hh++ )
						{
							WORD* pDstBits2 = (WORD*)pDstBits;

							for( ww = 0; ww < w; ww++ )
							{
								pDstBits2[ww] = GL_R8G8B8ToA1R5G5B5( Alpha, pPal[pBmpBits[ww]].rgbRed, pPal[pBmpBits[ww]].rgbGreen, pPal[pBmpBits[ww]].rgbBlue );
							}

							pBmpBits -= dwSrcRowBytes;
							pDstBits += dwDstRowBytes;
						}
					}
					break;
				case 24:

					for( hh = 0; hh < h; hh++ )
					{
						int s = 0;
						WORD* pDstBits2 = (WORD*)pDstBits;

						for( ww = 0; ww < w; ww++ )
						{
							pDstBits2[ww] = GL_R8G8B8ToA1R5G5B5( Alpha, pBmpBits[s+2], pBmpBits[s+1], pBmpBits[s+0] );
							s+=3;
						}

						pBmpBits -= dwSrcRowBytes;
						pDstBits += dwDstRowBytes;
					}

					break;
				case 32:

					for( hh = 0; hh < h; hh++ )
					{
						int s = 0;
						WORD* pDstBits2 = (WORD*)pDstBits;

						for( ww = 0; ww < w; ww++ )
						{
							pDstBits2[ww] = GL_R8G8B8ToA1R5G5B5( pBmpBits[s+3], pBmpBits[s+2], pBmpBits[s+1], pBmpBits[s+0] );
							s+=4;
						}

						pBmpBits -= dwSrcRowBytes;
						pDstBits += dwDstRowBytes;
					}

					break;
				default:

					ret = false;
				}
			}
			break;
		case GLFMT_X4R4G4B4:	// 16rbg 444
		case GLFMT_A4R4G4B4:
			{
				BYTE Alpha = ConvFormat==GLFMT_A4R4G4B4?0xFF:0;

				switch( pBmpInfo->bmiHeader.biBitCount )
				{
				case 8:
					{
						RGBQUAD* pPal = pBmpInfo->bmiColors;

						for( hh = 0; hh < h; hh++ )
						{
							WORD* pDstBits2 = (WORD*)pDstBits;

							for( ww = 0; ww < w; ww++ )
							{
								pDstBits2[ww] = GL_R8G8B8ToA4R4G4B4( Alpha, pPal[pBmpBits[ww]].rgbRed, pPal[pBmpBits[ww]].rgbGreen, pPal[pBmpBits[ww]].rgbBlue );
							}

							pBmpBits -= dwSrcRowBytes;
							pDstBits += dwDstRowBytes;
						}
					}
					break;
				case 24:

					for( hh = 0; hh < h; hh++ )
					{
						int s = 0;
						WORD* pDstBits2 = (WORD*)pDstBits;

						for( ww = 0; ww < w; ww++ )
						{
							pDstBits2[ww] = GL_R8G8B8ToA4R4G4B4( Alpha, pBmpBits[s+2], pBmpBits[s+1], pBmpBits[s+0] );
							s+=3;
						}

						pBmpBits -= dwSrcRowBytes;
						pDstBits += dwDstRowBytes;
					}

					break;
				case 32:

					for( hh = 0; hh < h; hh++ )
					{
						int s = 0;
						WORD* pDstBits2 = (WORD*)pDstBits;

						for( ww = 0; ww < w; ww++ )
						{
							pDstBits2[ww] = GL_R8G8B8ToA4R4G4B4( pBmpBits[s+3], pBmpBits[s+2], pBmpBits[s+1], pBmpBits[s+0] );
							s+=4;
						}

						pBmpBits -= dwSrcRowBytes;
						pDstBits += dwDstRowBytes;
					}

					break;
				default:

					ret = false;
				}
			}
			break;
		case GLFMT_R3G3B2:		// 8rbg 332

			switch( pBmpInfo->bmiHeader.biBitCount )
			{
			case 8:
				{
					RGBQUAD* pPal = pBmpInfo->bmiColors;

					for( hh = 0; hh < h; hh++ )
					{
						for( ww = 0; ww < w; ww++ )
						{
							pDstBits[ww] = GL_R8G8B8ToR3G3B2( pPal[pBmpBits[ww]].rgbRed, pPal[pBmpBits[ww]].rgbGreen, pPal[pBmpBits[ww]].rgbBlue );
						}

						pBmpBits -= dwSrcRowBytes;
						pDstBits += dwDstRowBytes;
					}
				}
				break;
			case 24:

				for( hh = 0; hh < h; hh++ )
				{
					int s = 0;

					for( ww = 0; ww < w; ww++ )
					{
						pDstBits[ww] = GL_R8G8B8ToR3G3B2( pBmpBits[s+2], pBmpBits[s+1], pBmpBits[s+0] );
						s+=3;
					}

					pBmpBits -= dwSrcRowBytes;
					pDstBits += dwDstRowBytes;
				}

				break;
			case 32:

				for( hh = 0; hh < h; hh++ )
				{
					int s = 0;

					for( ww = 0; ww < w; ww++ )
					{
						pDstBits[ww] = GL_R8G8B8ToR3G3B2( pBmpBits[s+2], pBmpBits[s+1], pBmpBits[s+0] );
						s+=4;
					}

					pBmpBits -= dwSrcRowBytes;
					pDstBits += dwDstRowBytes;
				}

				break;
			default:

				ret = false;
			}
			break;
		case GLFMT_A8R3G3B2:	// 16rbg 332
			{
				switch( pBmpInfo->bmiHeader.biBitCount )
				{
				case 8:
					{
						RGBQUAD* pPal = pBmpInfo->bmiColors;

						for( hh = 0; hh < h; hh++ )
						{
							WORD* pDstBits2 = (WORD*)pDstBits;

							for( ww = 0; ww < w; ww++ )
							{
								pDstBits2[ww] = GL_R8G8B8ToA8R3G3B2( 0xFF, pPal[pBmpBits[ww]].rgbRed, pPal[pBmpBits[ww]].rgbGreen, pPal[pBmpBits[ww]].rgbBlue );
							}

							pBmpBits -= dwSrcRowBytes;
							pDstBits += dwDstRowBytes;
						}
					}
					break;
				case 24:

					for( hh = 0; hh < h; hh++ )
					{
						int s = 0;
						WORD* pDstBits2 = (WORD*)pDstBits;

						for( ww = 0; ww < w; ww++ )
						{
							pDstBits2[ww] = GL_R8G8B8ToA8R3G3B2( 0xFF, pBmpBits[s+2], pBmpBits[s+1], pBmpBits[s+0] );
							s+=3;
						}

						pBmpBits -= dwSrcRowBytes;
						pDstBits += dwDstRowBytes;
					}

					break;
				case 32:

					for( hh = 0; hh < h; hh++ )
					{
						int s = 0;
						WORD* pDstBits2 = (WORD*)pDstBits;

						for( ww = 0; ww < w; ww++ )
						{
							pDstBits2[ww] = GL_R8G8B8ToA8R3G3B2( pBmpBits[s+3], pBmpBits[s+2], pBmpBits[s+1], pBmpBits[s+0] );
							s+=4;
						}

						pBmpBits -= dwSrcRowBytes;
						pDstBits += dwDstRowBytes;
					}

					break;
				default:

					ret = false;
				}
			}
			break;
		case GLFMT_P8:		// 8rbgpbg

			if( pBmpInfo->bmiHeader.biBitCount == 8 )
			{
				for( hh = 0; hh < h; hh++ )
				{
					::memmove( pDstBits, pBmpBits, w );

					pBmpBits -= dwSrcRowBytes;
					pDstBits += dwDstRowBytes;
				}
			}else
			{
				ret = false;
			}
			break;
		default:
			ret = false;
		}

	}
	::LocalUnlock( m_hBmpInfo );
	::LocalUnlock( m_hBmpBits );

	return ret;
}


/*----------------------------------------------------------------------------*
|	_O
*----------------------------------------------------------------------------*/
BOOL CGraphicLoader::Render( HDC hdc )
{
	if( m_hBmpInfo == NULL ) return FALSE;

	int ret;

	BITMAPINFO* pBmpInfo = (BITMAPINFO*)::LocalLock( m_hBmpInfo );
	BYTE* pBmpBits = (BYTE*)::LocalLock( m_hBmpBits );
	{

		if( !pBmpInfo || !pBmpBits ) return FALSE;

		ret = ::SetDIBitsToDevice( hdc, 0, 0, pBmpInfo->bmiHeader.biWidth, pBmpInfo->bmiHeader.biHeight, 0, 0, 0, 
			pBmpInfo->bmiHeader.biHeight, pBmpBits, pBmpInfo, DIB_RGB_COLORS );

	}
	::LocalUnlock( m_hBmpInfo );
	::LocalUnlock( m_hBmpBits );

	return ret;
}


#endif /* SLIB_USE_GRAPHICLOADER */